local auraDB = 0
local activeAuras = {}
local partyGUID = {['party1'] = 0, ['party2'] = 0, ['party3'] = 0, ['party4'] = 0}

function ShivvAudio:AudibleAuras()
	if (auraDB == 0) then
		auraDB = self.db.profile.auraDB -- local reference for speed (only update once)
	end

	-- clear out aura tracking entries
	for unit in pairs(activeAuras) do
		for auras in pairs(activeAuras[unit]) do
			activeAuras[unit][auras] = nil
		end
		activeAuras[unit] = nil
	end

	if (self.db.profile.auraEnableTracking) then
		-- aura tracking enabled, setup settings and register events
		for unit in pairs(auraDB) do
			-- create data table if doesn't already exist
			if (not activeAuras[unit]) then
				activeAuras[unit] = {}
			end

			-- silently scan unit (if it exists) to populate data table
			if (UnitExists(unit)) then
				self:ScanAuras(unit, true)
			end

			-- if the unit is a party member, record their gui for future changes
			if (strsub(unit, 1, 5) == ' party') then
				partyGUID[unit] = UnitGUID(unit)
			end
		end

		self:UNIT_PET(nil, 'player') -- force pet check on login

		-- register events
		self:RegisterEvent('UNIT_AURA')
		self:RegisterEvent('UNIT_PET')
		self:RegisterEvent('PLAYER_ENTERING_WORLD')
		self:RegisterEvent('PARTY_MEMBERS_CHANGED')
		self:RegisterEvent('PLAYER_TARGET_CHANGED')
		self:RegisterEvent('PLAYER_FOCUS_CHANGED')
	else
		-- aura tracking disabled, cleanout data and unregister events
		self:UnregisterEvent('UNIT_AURA')
		self:UnregisterEvent('UNIT_PET')
		self:UnregisterEvent('PLAYER_FOCUS_CHANGED')
		self:UnregisterEvent('PLAYER_TARGET_CHANGED')
		self:UnregisterEvent('PLAYER_ENTERING_WORLD')
		self:UnregisterEvent('PARTY_MEMBERS_CHANGED')

		self:CleanupMusic()
	end
end

-- ------------ --
-- AURA SCANNER --
-- ------------ --
do
	local GetTime, UnitBuff, UnitDebuff = _G.GetTime, _G.UnitBuff, _G.UnitDebuff -- local references for speed
	local aura, i, time, unitData, activeData

	function ShivvAudio:ScanAuras(unit, silent)
		unitData, activeData = auraDB[unit], activeAuras[unit] -- local references for speed
		time = GetTime()

		-- scan buffs on unit for tracked auras
		i, aura = 1, UnitBuff(unit, 1)

		while(aura) do
			if (unitData[aura]) then -- found a buff we are tracking
				if (activeData[aura]) then -- already being tracked, update timestamp
					activeData[aura] = time
				else -- not being tracked, its new!
					activeData[aura] = time

					if (not silent and unitData[aura].onGain ~= 0) then -- this isn't a silent scan and an onGain sound exists
						self:SoundHandler(unitData[aura].onGain)
					end

					if (unit == 'player' and unitData[aura].music ~= 0) then -- scanning player and this aura has music set, play it
						self:MusicHandler(aura, true)
					end
				end
			end

			i = i + 1
			aura = UnitBuff(unit, i)
		end

		-- now do the exact same thing but for debuffs this time, ugh
		i, aura = 1, UnitDebuff(unit, 1)

		while(aura) do
			if (unitData[aura]) then -- found a buff we are tracking
				if (activeData[aura]) then -- already being tracked, update timestamp
					activeData[aura] = time
				else -- not being tracked, its new!
					activeData[aura] = time

					if (not silent and unitData[aura].onGain ~= 0) then -- this isn't a silent scan and an onGain sound exists
						self:SoundHandler(unitData[aura].onGain)
					end

					if (unit == 'player' and unitData[aura].music ~= 0) then -- scanning player and this aura has music set, play it
						self:MusicHandler(aura, true)
					end
				end
			end

			i = i + 1
			aura = UnitDebuff(unit, i)
		end

		-- we've found all new buffs, lets get rid of any old ones
		for aura, timestamp in pairs(activeData) do
			if (timestamp ~= time) then -- aura is no longer active, remove it
				activeData[aura] = nil

				if (not silent and unitData[aura].onLoss ~= 0) then -- this isn't a silent scan and an onLoss sound exists
					self:SoundHandler(unitData[aura].onLoss)
				end

				if (unit == 'player' and unitData[aura].music ~= 0) then -- scanning player and this aura had music set, kill it
					self:MusicHandler(aura, false)
				end
			end
		end
	end
end

-- -------------- --
-- EVENT HANDLERS --
-- -------------- --
function ShivvAudio:PLAYER_ENTERING_WORLD()
	self:StartDelayedActionTimer(0.5, 'MusicForcePlay')
end

function ShivvAudio:UNIT_AURA(event, unit)
	if (activeAuras[unit]) then
		self:ScanAuras(unit)
	end
end

function ShivvAudio:PLAYER_FOCUS_CHANGED()
	if (activeAuras['focus'] and UnitExists('focus')) then -- do something if we're tracking the focus and it exists
		self:ScanAuras('focus', true) -- scan silently
	end
end

function ShivvAudio:PLAYER_TARGET_CHANGED()
	if (activeAuras['target'] and UnitExists('target')) then -- do something if we're tracking the target and it exists
		self:ScanAuras('target', true) -- scan silently
	end
end

function ShivvAudio:PARTY_MEMBERS_CHANGED()
	for unit in pairs(partyGUID) do
		if (activeAuras[unit] and UnitExists(unit)) then -- if we're tracking this party member and they exist
			if (partyGUID ~= UnitGUID(unit)) then -- either a different party member or is one where there wasn't before
				self:ScanAuras(unit, true) -- scan silently
				partyGUID[unit] = UnitGUID(unit)
			end
		else -- no party member here
			partyGUID[unit] = 0
		end
	end
end

do
	local petGUID = 0

	function ShivvAudio:UNIT_PET(event, unit)
		if (unit ~= 'player') then return end

		if (activeAuras['pet'] and UnitExists('pet')) then
			if (petGUID ~= UnitGUID('pet')) then -- either a different pet or have one when didn't before
				self:ScanAuras('pet', true) -- scan silently
				petGUID = UnitGUID('pet')
			end
		else -- pet doesn't exist, clear guid
			petGUID = 0
		end
	end
end