ShivvAudio = LibStub('AceAddon-3.0'):NewAddon('ShivvAudio', 'AceConsole-3.0', 'AceEvent-3.0')

local mediaDB, auraDB, eventDB
local defaults = {
	profile = {
		-- global options
		auraEnableTracking = false,
		vitalEnableHealthTracking = false,
		vitalEnablePowerTracking = false,
		gameMusicControl = false,
		-- aura data
		auraDB = {},
		-- vital data
		vitalHealthDB = {
			['player'] = {['enable'] = false, ['sound'] = 0, ['percent'] = 0.5},
			['pet'] = {['enable'] = false, ['sound'] = 0, ['percent'] = 0.5},
			['focus'] = {['enable'] = false, ['sound'] = 0, ['percent'] = 0.5},
			['target'] = {['enable'] = false, ['sound'] = 0, ['percent'] = 0.5},
			['party1'] = {['enable'] = false, ['sound'] = 0, ['percent'] = 0.5},
			['party2'] = {['enable'] = false, ['sound'] = 0, ['percent'] = 0.5},
			['party3'] = {['enable'] = false, ['sound'] = 0, ['percent'] = 0.5},
			['party4'] = {['enable'] = false, ['sound'] = 0, ['percent'] = 0.5},
			['arena1'] = {['enable'] = false, ['sound'] = 0, ['percent'] = 0.5},
			['arena2'] = {['enable'] = false, ['sound'] = 0, ['percent'] = 0.5},
			['arena3'] = {['enable'] = false, ['sound'] = 0, ['percent'] = 0.5},
			['arena4'] = {['enable'] = false, ['sound'] = 0, ['percent'] = 0.5},
			['arena5'] = {['enable'] = false, ['sound'] = 0, ['percent'] = 0.5},
		},
		vitalPowerDB = {
			['player'] = {['enable'] = false, ['sound'] = 0, ['percent'] = 0.5},
			['pet'] = {['enable'] = false, ['sound'] = 0, ['percent'] = 0.5},
			['focus'] = {['enable'] = false, ['sound'] = 0, ['percent'] = 0.5},
			['target'] = {['enable'] = false, ['sound'] = 0, ['percent'] = 0.5},
			['party1'] = {['enable'] = false, ['sound'] = 0, ['percent'] = 0.5},
			['party2'] = {['enable'] = false, ['sound'] = 0, ['percent'] = 0.5},
			['party3'] = {['enable'] = false, ['sound'] = 0, ['percent'] = 0.5},
			['party4'] = {['enable'] = false, ['sound'] = 0, ['percent'] = 0.5},
			['arena1'] = {['enable'] = false, ['sound'] = 0, ['percent'] = 0.5},
			['arena2'] = {['enable'] = false, ['sound'] = 0, ['percent'] = 0.5},
			['arena3'] = {['enable'] = false, ['sound'] = 0, ['percent'] = 0.5},
			['arena4'] = {['enable'] = false, ['sound'] = 0, ['percent'] = 0.5},
			['arena5'] = {['enable'] = false, ['sound'] = 0, ['percent'] = 0.5},
		},
		vitalRepeatTimer = 5,
		-- event data
		eventDB = {
			-- sounds
			['death'] = 0,
			['levelup'] = 0,
			['achievement'] = 0,
			['combatstart'] = 0,
			['combatstop'] = 0,
			-- music
			['shop'] = 0,
			['auction'] = 0,
			['bank'] = 0,
		},
	},
	global = {
		mediaDB = {},
	},
}

-- ------------------- --
-- DELAYED ACTION TIMER
-- -------------------- --
do
	local handler = CreateFrame('Frame')
	local timer, timerFunc

	local function DelayedActionTimer(frame, elapsed)
		timer = timer - elapsed

		if (timer <= 0) then
			handler:SetScript('OnUpdate', nil)
			ShivvAudio[timerFunc](ShivvAudio)
		end
	end

	function ShivvAudio:StartDelayedActionTimer(time, func)
		timer, timerFunc = time, func
		handler:SetScript('OnUpdate', DelayedActionTimer)
	end
end

-- -------- --
-- MOD INIT --
-- -------- --
function ShivvAudio:OnInitialize()
	self.db = LibStub('AceDB-3.0'):New('ShivvAudioDB', defaults)
	self:RegisterChatCommand('shivv', 'ChatHandler')

	mediaDB, auraDB, eventDB = self.db.global.mediaDB, self.db.profile.auraDB, self.db.profile.eventDB -- local references for speed

	-- do something here about version checking
end

function ShivvAudio:OnEnable()
	self:MusicalEvents()
	self:StartDelayedActionTimer(0.5, 'DelayedOnEnable')
end

function ShivvAudio:DelayedOnEnable()
	self:AudibleAuras()
	self:MelodicVitals(true, true)
end

-- ------------- --
-- SOUND HANDLER --
-- ------------- --
function ShivvAudio:SoundHandler(sound)
	PlaySoundFile(mediaDB[sound].path)
end

-- ------------- --
-- MUSIC HANDLER --
-- ------------- --
do
	local activeMusic = {}
	local internalMusicRef = {['shivvInternalAuction'] = 'auction', ['shivvInternalBank'] = 'bank',	['shivvInternalShop'] = 'shop'}

	function ShivvAudio:CleanupMusic()
		-- clear out and kill any active music
		while (#(activeMusic) > 0) do
			tremove(activeMusic, 1)
		end

		self:MusicStop()
	end

	function ShivvAudio:MusicForcePlay()
		-- force play the #1 music (if any)
		if (#(activeMusic) >= 1) then
			self:MusicStart(activeMusic[1])
		end
	end

	function ShivvAudio:IsMusicActive()
		if (#(activeMusic) > 0) then
			return true
		else
			return false
		end
	end

	function ShivvAudio:MusicStart(aura)
		if (self.db.profile.gameMusicControl and GetCVar('Sound_EnableMusic') == '0') then -- we control music and its off, turn it on
			SetCVar('Sound_EnableMusic', '1')
		end

		if (internalMusicRef[aura]) then -- this is a musical event music
			PlayMusic(mediaDB[eventDB[internalMusicRef[aura]]].path)
		else -- this is an audible aura music
			PlayMusic(mediaDB[auraDB.player[aura].music].path)
		end
	end

	function ShivvAudio:MusicStop()
		StopMusic()

		if (self.db.profile.gameMusicControl and GetCVar('Sound_EnableMusic') == '1') then -- we control music and its on, kill it
			SetCVar('Sound_EnableMusic', '0')
		end
	end

	function ShivvAudio:MusicHandler(aura, active)
		if (active) then -- new music to activate, do so and add to tracking table
			self:MusicStart(aura)

			tinsert(activeMusic, 1, aura)
		else -- music to stop plauying, do so, remove from tracking table and look for replacement (if any)
			for index, active in ipairs(activeMusic) do
				if (aura == active) then -- found this aura in active data
					if (index ~= 1) then -- this wasn't currently playing, just remove entry
						tremove(activeMusic, index)
					else -- this is currently playing, remove entry and look for replacement
						tremove(activeMusic, index)

						if (#(activeMusic) >= 1) then -- there is another entry, play that instead
							self:MusicStart(activeMusic[1])
						else -- no other entry, stop music
							self:MusicStop()
						end
					end

					break -- in any case, we're done here
				end
			end
		end
	end
end
