local G = LibStub("AceGUI-3.0")
local L = LibStub('AceLocale-3.0'):GetLocale('ShivvAudio')

local LSM -- LibSharedMedia if available
local addonMediaPath = 'Interface\\AddOns\\ShivvAudio\\Media\\'
local mediaListDropdown = {}
local tabTable = {
	[1] = {text = L.Global, value = 'global'},
	[2] = {text = L.MediaDatabase, value = 'media'},
	[3] = {text = L.AudibleAuras, value = 'auras'},
	[4] = {text = L.MelodicVitals, value = 'vitals'},
	[5] = {text = L.MusicalEvents, value = 'events'},
}
local unitListDropdown = {
	['player'] = L.Player,
	['pet'] = L.Pet,
	['focus'] = L.Focus,
	['target'] = L.Target,
	['party1'] = L.Party1,
	['party2'] = L.Party2,
	['party3'] = L.Party3,
	['party4'] = L.Party4,
	['arena1'] = L.Arena1,
	['arena2'] = L.Arena2,
	['arena3'] = L.Arena3,
	['arena4'] = L.Arena4,
	['arena5'] = L.Arena5,
}

local function MakeItYellow(text) -- stupid but amuses me (and does mean less coding!)
	return '|cffffd100' .. text .. '|r'
end

local function ZeroToNone(v)
	if (v == 0) then
		return 'none'
	else
		return v
	end
end

local function NoneToZero(v)
	if (v == 'none') then
		return 0
	else
		return v
	end
end

local function Spacer(w, h)
	local spacer = G:Create('Label')
	spacer:SetText('')
	spacer:SetWidth(w)
	spacer:SetHeight(h)
	return spacer
end

local function ShivvTip(widget, title, text)
	GameTooltip:SetOwner(widget.frame, 'ANCHOR_TOPRIGHT')
	GameTooltip:SetText(title, 1, 1, 1, 1, 1)
	GameTooltip:AddLine(text, 1, 0.83, 0, 1, 1)
	GameTooltip:Show()
end

-- ------------------- --
-- MEDIA LIST DROPDOWN --
-- ------------------- --
do
	local sortedList = {}
	local tsort = table.sort
	local mDB

	local function MakeItBlue(text)
		return '|cff40bad8' .. text .. '|r'
	end

	function ShivvAudio:PopulateMediaListDropdown()
		for k in pairs(mediaListDropdown) do mediaListDropdown[k] = nil end -- clear list

		mDB = self.db.global.mediaDB

		for k in pairs(mDB) do
			sortedList[#sortedList + 1] = k
		end
		tsort(sortedList)

		for i, key in ipairs(sortedList) do
			if (mDB[key].isMusic) then
				mediaListDropdown[key] = MakeItBlue(key)
			else
				mediaListDropdown[key] = key
			end
			sortedList[i] = nil
		end

		mediaListDropdown['none'] = L.None -- add 'none' entry
	end
end

-- --------- --
-- BUILD GUI --
-- --------- --
function ShivvAudio:BuildGUI()
	if (self:IsMusicActive()) then -- don't open gui unless all 'musical' buffs are cancelled (avoids a lot of problems)
		self:TextOutput(L.NoGUIWhileMusic)
		return
	end

	-- create a new AceGUI layout to avoid a bug with positioning two side-by-side-frames
	G:RegisterLayout('KithTwoPanel', function(content, children)
		if (children[1]) then
			local child = children[1]
			child.frame:ClearAllPoints()
			child.frame:Show()
			child.frame:SetPoint('TOPLEFT', content, 'TOPLEFT', 0, -5)
		end
		if (children[2]) then
			local child = children[2]
			child.frame:ClearAllPoints()
			child.frame:Show()
			child.frame:SetPoint('TOPRIGHT', content, 'TOPRIGHT', 0, -5)
		end
	end)

	local f = G:Create('Frame')
	f:SetTitle(L.Title .. '  |cffffffffv' .. GetAddOnMetadata('ShivvAudio', 'Version') .. '|r')
	f:SetStatusText(L.StatusText)
	f:SetCallback('OnClose', function(widget) G:Release(widget) end)
	f:SetLayout('Fill')
	f:Show()

	local t = G:Create('TabGroup')
	t:SetTitle('')
	t:SetTabs(tabTable)
	t:SetLayout('Flow')
	t:SetCallback('OnGroupSelected', function(group, event, value)
		group:ReleaseChildren()

		if (value == 'media') then
			self:ShowMedia(group)
		elseif (value == 'auras') then
			self:ShowAuras(group)
		elseif (value == 'vitals') then
			self:ShowVitals(group)
		elseif (value == 'events') then
			self:ShowEvents(group)
		else
			self:ShowGlobal(group)
		end
	end)
	t:SelectTab('global')
	f:AddChild(t)
end

-- ------------ --
-- GLOBAL PANEL --
-- ------------ --
function ShivvAudio:ShowGlobal(f)
	local full = G:Create('SimpleGroup')
	full:SetLayout('KithTwoPanel')
	full:SetFullHeight(true)
	full:SetFullWidth(true)
	f:AddChild(full)

	local opt = G:Create('SimpleGroup')
	opt:SetLayout('List')
	opt:SetWidth(220)
	opt:SetFullHeight(true)
	full:AddChild(opt)

	local eAura = G:Create('CheckBox')
	eAura:SetValue(self.db.profile.auraEnableTracking)
	eAura:SetTriState(false)
	eAura:SetType('checkbox')
	eAura:SetLabel(MakeItYellow(L.EnableAura))
	eAura:SetDescription(L.EnableAuraDesc)
	eAura:SetCallback('OnValueChanged', function(w, e, v)
		self.db.profile.auraEnableTracking = v
		self:AudibleAuras()
	end)
	opt:AddChild(eAura)

	local vHealth = G:Create('CheckBox')
	vHealth:SetValue(self.db.profile.vitalEnableHealthTracking)
	vHealth:SetTriState(false)
	vHealth:SetType('checkbox')
	vHealth:SetLabel(MakeItYellow(L.EnableHealth))
	vHealth:SetDescription(L.EnableHealthDesc)
	vHealth:SetCallback('OnValueChanged', function(w, e, v)
		self.db.profile.vitalEnableHealthTracking = v
		self:MelodicVitals(true, false)
	end)
	opt:AddChild(vHealth)

	local vPower = G:Create('CheckBox')
	vPower:SetValue(self.db.profile.vitalEnablePowerTracking)
	vPower:SetTriState(false)
	vPower:SetType('checkbox')
	vPower:SetLabel(MakeItYellow(L.EnablePower))
	vPower:SetDescription(L.EnablePowerDesc)
	vPower:SetCallback('OnValueChanged', function(w, e, v)
		self.db.profile.vitalEnablePowerTracking = v
		self:MelodicVitals(false, true)
	end)
	opt:AddChild(vPower)

	local music = G:Create('CheckBox')
	music:SetValue(self.db.profile.gameMusicControl)
	music:SetTriState(false)
	music:SetType('checkbox')
	music:SetLabel(MakeItYellow(L.EnableMusicControl))
	music:SetDescription(L.EnableMusicControlDesc)
	music:SetCallback('OnValueChanged', function(w, e, v)
		self.db.profile.gameMusicControl = v
	end)
	opt:AddChild(music)

	local timer = G:Create('Slider')
	timer:SetSliderValues(1, 20, 1)
	timer:SetIsPercent(false)
	timer:SetWidth(200)
	timer:SetLabel('Vital Announce Reset')
	timer:SetValue(self.db.profile.vitalRepeatTimer)
	timer:SetCallback('OnEnter', function(w)
		ShivvTip(w, L.VitalAnnounceTip, L.VitalAnnounceDesc)
	end)
	timer:SetCallback('OnLeave', function(w)
		GameTooltip:Hide()
	end)
	timer:SetCallback('OnValueChanged', function(w, e, v)
		self.db.profile.vitalRepeatTimer = math.floor(v + 0.5) -- round to nearest 1
		self:VitalEventSetup()
	end)
	opt:AddChild(timer)

	opt:AddChild(Spacer(100, 8))

	-- setup LSM import buton
	local lsm = G:Create('Button')
	lsm:SetWidth(200)

	LSM = LibStub:GetLibrary('LibSharedMedia-3.0', true)

	if (LSM) then -- LSM installed, activate button
		lsm:SetText(L.AddLSM)
		lsm:SetDisabled(false)
		lsm:SetCallback('OnClick', function(widget)
			self:LSMImport()
		end)
		lsm:SetCallback('OnEnter', function(widget)
			ShivvTip(widget, L.AddLSM, L.AddLSMDescEnabled)
		end)
		lsm:SetCallback('OnLeave', function(widget)
			GameTooltip:Hide()
		end)
	else -- no LSM installed, disable button
		lsm:SetText(L.AddLSMDisabled)
		lsm:SetDisabled(true)
	end
	opt:AddChild(lsm)

	local introBox = G:Create('InlineGroup')
	introBox:SetTitle(L.IntroTitle)
	introBox:SetWidth(400)
	introBox:SetLayout('Flow')
	full:AddChild(introBox)

	local introScroll = G:Create('ScrollFrame')
	introScroll:SetLayout('List')
	introScroll:SetRelativeWidth(1)
	introScroll:SetHeight(340)
	introBox:AddChild(introScroll)

	local intro = G:Create('Label')
	intro:SetText(L.Intro)
	intro:SetColor(1, 1, 1)
	intro:SetFontObject(GameFontNormalSmall)
	intro:SetRelativeWidth(1)
	introScroll:AddChild(intro)
end

-- ----------- --
-- MEDIA PANEL --
-- ----------- --
do
	local sortedList = {}
	local tsort = table.sort
	local item, data, frame, scroll, addEdit

	local function IsItMusic(entry) -- colours musical tagged entries blue for reference
		if (entry) then
			return '|cff40bad8' .. L.Yes .. '|r'
		else
			return L.No
		end
	end

	local function PopulateMediaDatabase(self)
		-- clear data
		scroll:ReleaseChildren()
		scroll:SetUserData('selectedMedia', nil)
		scroll:SetUserData('selectedItemRef', nil)

		-- create an alphabetically sorted list of media
		for media in pairs(self.db.global.mediaDB) do
			sortedList[#sortedList + 1] = media
		end
		tsort(sortedList)

		-- create scrollframe entries for each media item
		for i, key in ipairs(sortedList) do
			data = self.db.global.mediaDB[key] -- local lookup reference
			item = G:Create('PentaTextInteractiveLabel')
			item:SetRelativeWidth(1)
			item:SetHighlight('Interface\\QuestFrame\\UI-QuestTitleHighlight')
			item:SetTextWidths(150, 60, 0, 0)
			item:SetColor()
			item:SetText(key, IsItMusic(data.isMusic), nil, nil, data.path)
			item:SetUserData('media', key)
			item:SetCallback('OnClick', function(w)
				if (w:IsSelected()) then -- already selected, deselect
					scroll:SetUserData('selectedMedia', nil)
					scroll:SetUserData('selecteItemRef', nil)
					ShivvAudio:ShowMediaAddEdit(true)
				else
					if (scroll:GetUserData('selectedItemRef') ~= nil) then -- one already selected, clear that before selecting self
						scroll:GetUserData('selectedItemRef'):UnlockHighlight()
					end

					scroll:SetUserData('selectedMedia', w:GetUserData('media'))
					scroll:SetUserData('selectedItemRef', w)
					ShivvAudio:ShowMediaAddEdit(false, w:GetUserData('media'))
				end
			end)
			scroll:AddChild(item)
			sortedList[i] = nil
		end
	end

	function ShivvAudio:ShowMedia(f)
		frame = f -- for later external ref

		local box = G:Create('InlineGroup')
		box:SetTitle(L.ListOfMedia)
		box:SetRelativeWidth(1)
		box:SetLayout('Flow')
		f:AddChild(box)

		local header = G:Create('PentaTextInteractiveLabel')
		header:SetRelativeWidth(1)
		header:SetHighlight(nil)
		header:SetColor(1, 0.83, 0, 1, 0.83, 0, nil, nil, nil, nil, nil, nil, 1, 0.83, 0)
		header:SetTextWidths(150, 60, 0, 0)
		header:SetText(L.MediaHeaderName, L.MediaHeaderMusic, nil, nil, L.MediaHeaderPath)
		header:SetFontObject(GameFontNormalSmall)
		box:AddChild(header)

		local s = G:Create('ScrollFrame')
		s:SetLayout('List')
		s:SetRelativeWidth(1)
		s:SetHeight(150)
		box:AddChild(s)

		scroll = s -- for later external ref

		PopulateMediaDatabase(self)

		local ae = G:Create('InlineGroup')
		ae:SetRelativeWidth(1)
		ae:SetLayout('Flow')
		f:AddChild(ae)
		addEdit = ae

		self:ShowMediaAddEdit(true)
	end

	function ShivvAudio:ShowMediaAddEdit(addMode, key)
		addEdit:ReleaseChildren()

		if (addMode) then
			addEdit:SetTitle(L.MediaAdd)
		else
			addEdit:SetTitle(L.MediaEdit)
		end

		local name = G:Create('EditBox')
		name:SetLabel(L.MediaName)
		name:SetWidth(250)
		if (not addMode) then
			name:SetText(key)
			name:SetDisabled(true)
		else
			name:SetDisabled(false)
		end
		name:SetCallback('OnEnter', function(w)
			ShivvTip(w, L.MediaNameTip, L.MediaNameDesc)
		end)
		name:SetCallback('OnLeave', function()
			GameTooltip:Hide()
		end)
		addEdit:AddChild(name)

		local music = G:Create('CheckBox')
		music:SetTriState(false)
		music:SetType('checkbox')
		music:SetWidth(100)
		music:SetLabel(L.MediaIsMusic)
		if (addMode) then
			music:SetValue(false)
		else
			music:SetValue(self.db.global.mediaDB[key].isMusic)
		end
		music:SetCallback('OnEnter', function(w)
			ShivvTip(w, L.MediaIsMusicTip, L.MediaIsMusicDesc)
		end)
		music:SetCallback('OnLeave', function()
			GameTooltip:Hide()
		end)
		addEdit:AddChild(music)

		local path = G:Create('EditBox')
		path:SetLabel(L.MediaPath)
		path:SetWidth(250)
		if (addMode) then
			path:SetCallback('OnEnter', function(w)
				ShivvTip(w, L.MediaPathTip, L.MediaPathDesc)
			end)
		else
			path:SetText(self.db.global.mediaDB[key].path)
			path:SetCallback('OnEnter', function(w)
				ShivvTip(w, L.MediaPathTip, L.MediaPathDescEdit)
			end)
		end
		path:SetCallback('OnLeave', function()
			GameTooltip:Hide()
		end)
		addEdit:AddChild(path)

		if (addMode) then
			local internal = G:Create('CheckBox')
			internal:SetTriState(false)
			internal:SetType('checkbox')
			internal:SetWidth(160)
			internal:SetLabel(L.MediaInShivv)
			internal:SetValue(false)
			internal:SetCallback('OnEnter', function(w)
				ShivvTip(w, L.MediaInShivvTip, L.MediaInShivvDesc)
			end)
			internal:SetCallback('OnLeave', function()
				GameTooltip:Hide()
			end)
			addEdit:AddChild(internal)

			local addIt = G:Create('Button')
			addIt:SetText(L.MediaAddIt)
			addIt:SetWidth(140)
			addIt:SetCallback('OnClick', function(w)
				if (name:GetText() == '' or path:GetText() == '' or self.db.global.mediaDB[name:GetText()]) then
					self:TextOutput(L.MediaMustEnterUniqueNameAndPath)
				else
					local n, p, m, i = name:GetText(), path:GetText(), music:GetValue(), internal:GetValue()

					if (i) then p = addonMediaPath .. p end -- prepend shivvaudio\media path if is internal sound

					-- create new media entry
					self.db.global.mediaDB[n] = {
						['isMusic'] = m,
						['path'] = p,
					}

					PopulateMediaDatabase(self)
					self:TextOutput(L.NewMediaAdded)
					self:ShowMediaAddEdit(true)
				end
			end)
			addEdit:AddChild(addIt)
		else
			local sg = G:Create('SimpleGroup')
			sg:SetLayout('Flow')
			sg:SetRelativeWidth(1)
			addEdit:AddChild(sg)

			local editIt = G:Create('Button')
			editIt:SetText(L.MediaEditIt)
			editIt:SetWidth(140)
			editIt:SetCallback('OnClick', function(w)
				if (path:GetText() == '') then -- checks for a path entry
					self:TextOutput(L.MediaMustEnterPath)
					path:SetText(self.db.global.mediaDB[key].path)
				else
					local p, m = path:GetText(), music:GetValue()

					-- update entry
					self.db.global.mediaDB[key].isMusic = m
					self.db.global.mediaDB[key].path = p

					PopulateMediaDatabase(self)
					self:TextOutput(L.EditedMedia)
					self:ShowMediaAddEdit(true) -- hide edit box as we've deselected the edited option
				end
			end)
			addEdit:AddChild(editIt)

			local delIt = G:Create('Button')
			delIt:SetText(L.MediaDelIt)
			delIt:SetWidth(140)
			delIt:SetCallback('OnClick', function(w)
				self:RemoveMedia(key)

				PopulateMediaDatabase(self)
				self:TextOutput(L.DeletedMedia)
				self:ShowMediaAddEdit(true)
			end)
			addEdit:AddChild(delIt)

			local play = G:Create('Button')
			play:SetText(L.PlayMedia)
			play:SetWidth(140)
			play:SetCallback('OnClick', function(w)
				self:SoundHandler(key)
				self:TextOutput(L.PlayingMedia)
			end)
			addEdit:AddChild(play)
		end
	end
end

-- ------------ --
-- MEDIA DELETE --
-- ------------ --
function ShivvAudio:RemoveMedia(media)
	local changed = 0

	self:TextOutput(L.RemovingMedia)

	for char, charData in pairs(ShivvAudioDB.profiles) do
		self:TextOutput(L.Scanning .. char)

		-- first: audible auras
		if (charData.auraDB) then
			changed = 0
			for unit in pairs(charData.auraDB) do
				for buff in pairs(charData.auraDB[unit]) do
					if (charData.auraDB[unit][buff].onGain == media) then
						charData.auraDB[unit][buff].onGain = 0
						changed = changed + 1
					end
					if (charData.auraDB[unit][buff].onLoss == media) then
						charData.auraDB[unit][buff].onLoss = 0
						changed = changed + 1
					end
					if (charData.auraDB[unit][buff].music == media) then
						charData.auraDB[unit][buff].music = 0
						changed = changed + 1
					end
				end
			end

			if (changed > 0) then
				self:TextOutput(L.RemoveChanged .. changed .. L.RemoveAura)
			end
		end

		-- then melodic vitals
		if (charData.vitalHealthDB) then
			changed = 0
			for unit in pairs(charData.vitalHealthDB) do
				if (charData.vitalHealthDB[unit].sound == media) then
					charData.vitalHealthDB[unit].sound = 0
					charData.vitalHealthDB[unit].enable = false
					changed = changed + 1
				end
			end

			if (changed > 0) then
				self:TextOutput(L.RemoveChanged .. changed .. L.RemoveVitalHP)
			end
		end

		if (charData.vitalPowerDB) then
			changed = 0
			for unit in pairs(charData.vitalPowerDB) do
				if (charData.vitalPowerDB[unit].sound == media) then
					charData.vitalPowerDB[unit].sound = 0
					charData.vitalPowerDB[unit].enable = false
					changed = changed + 1
				end
			end

			if (changed > 0) then
				self:TextOutput(L.RemoveChanged .. changed .. L.RemoveVitalPWR)
			end
		end

		-- then musical events
		if (charData.eventDB) then
			changed = 0

			if (charData.eventDB.death and charData.eventDB.death == media) then
				charData.eventDB.death = 0
				changed = changed + 1
			end

			if (charData.eventDB.levelup and charData.eventDB.levelup == media) then
				charData.eventDB.levelup = 0
				changed = changed + 1
			end

			if (charData.eventDB.achievement and charData.eventDB.achievement == media) then
				charData.eventDB.achievement = 0
				changed = changed + 1
			end

			if (charData.eventDB.shop and charData.eventDB.shop == media) then
				charData.eventDB.shop = 0
				changed = changed + 1
			end

			if (charData.eventDB.auction and charData.eventDB.auction == media) then
				charData.eventDB.auction = 0
				changed = changed + 1
			end

			if (charData.eventDB.bank and charData.eventDB.bank == media) then
				charData.eventDB.bank = 0
				changed = changed + 1
			end

			if (changed > 0) then
				self:TextOutput(L.RemoveChanged .. changed .. L.RemoveEvents)
			end
		end
	end

	-- reset any events/tracking for these modules on the current player
	self:MelodicVitals(true, true)
	self:MusicalEvents()

	-- and finally delete the actual entry
	self.db.global.mediaDB[media] = nil
end

-- ---------- --
-- AURA PANEL --
-- ---------- --
do
	local unitList, buffList, interimList = {}, {}, {}
	local tsort = table.sort
	local item, frame, scroll, data, addEdit
	local sortedBy = 'unit'

	local function ShowNiceValue(v)
		if (v == 0) then
			return '--'
		else
			return v
		end
	end

	local function sizeOf(tbl)
		local n = 0
		for _ in pairs(tbl) do
			n = n + 1
		end

		return n
	end

	local function AddAuraDatabaseItem(unit, buff, onGain, onLoss, music)
		item = G:Create('PentaTextInteractiveLabel')
		item:SetRelativeWidth(1)
		item:SetHighlight('Interface\\QuestFrame\\UI-QuestTitleHighlight')
		item:SetTextWidths(70, 170, 120, 120)
		item:SetColor()
		item:SetText(unit, buff, ShowNiceValue(onGain), ShowNiceValue(onLoss), ShowNiceValue(music))
		item:SetUserData('unit', unit)
		item:SetUserData('buff', buff)
		item:SetCallback('OnClick', function(w)
			if (w:IsSelected()) then
				scroll:SetUserData('selectedUnit', nil)
				scroll:SetUserData('selectedBuff', nil)
				scroll:SetUserData('selectedItemRef', nil)
				ShivvAudio:ShowAuraAddEdit(true)
			else
				if (scroll:GetUserData('selectedItemRef') ~= nil) then
					scroll:GetUserData('selectedItemRef'):UnlockHighlight()
				end

				scroll:SetUserData('selectedUnit', w:GetUserData('unit'))
				scroll:SetUserData('selectedBuff', w:GetUserData('buff'))
				scroll:SetUserData('selectedItemRef', w)
				ShivvAudio:ShowAuraAddEdit(false, w:GetUserData('unit'), w:GetUserData('buff'))
			end
		end)
		scroll:AddChild(item)
	end

	local function PopulateAuraDatabase(self)
		scroll:ReleaseChildren()
		scroll:SetUserData('selectedUnit', nil)
		scroll:SetUserData('selectedBuff', nil)
		scroll:SetUserData('selectedItemRef', nil)

		if (sortedBy == 'unit') then
			for unit in pairs(self.db.profile.auraDB) do
				unitList[#unitList + 1] = unit
			end
			tsort(unitList)

			for i, unit in ipairs(unitList) do
				for buff in pairs(self.db.profile.auraDB[unit]) do
					buffList[#buffList + 1] = buff
				end
				tsort(buffList)

				for j, buff in ipairs(buffList) do
					data = self.db.profile.auraDB[unit][buff]

					AddAuraDatabaseItem(unit, buff, data.onGain, data.onLoss, data.music)

					buffList[j] = nil
				end

				unitList[i] = nil
			end
		else
			-- create a 'new' auraDB that section buf buff first, then unit
			for unit in pairs(self.db.profile.auraDB) do
				for buff in pairs(self.db.profile.auraDB[unit]) do
					if (not interimList[buff]) then
						interimList[buff] = {}
					end

					interimList[buff][unit] = true
				end
			end

			for buff in pairs(interimList) do
				buffList[#buffList + 1] = buff
			end
			tsort(buffList)

			for i, buff in ipairs(buffList) do
				for unit in pairs(interimList[buff]) do
					unitList[#unitList + 1] = unit
				end
				tsort(unitList)

				for j, unit in ipairs(unitList) do
					data = self.db.profile.auraDB[unit][buff]
					AddAuraDatabaseItem(unit, buff, data.onGain, data.onLoss, data.music)
					unitList[j] = nil
				end
				buffList[i] = nil
			end

			-- clean out interim table
			for buff in pairs(interimList) do
				for unit in pairs(interimList[buff]) do
					interimList[buff][unit] = nil
				end
				interimList[buff] = nil
			end
		end
	end

	function ShivvAudio:ShowAuras(f)
		frame = f

		local box = G:Create('InlineGroup')
		box:SetTitle(L.ListOfAuras)
		box:SetRelativeWidth(1)
		box:SetLayout('Flow')
		f:AddChild(box)

		local header = G:Create('PentaTextInteractiveLabel')
		header:SetRelativeWidth(1)
		header:SetHighlight(nil)
		header:SetColor(1, 0.83, 0, 1, 0.83, 0, 1, 0.83, 0, 1, 0.83, 0, 1, 0.83, 0)
		header:SetTextWidths(70, 170, 120, 120)
		header:SetText(L.AuraHeaderUnit, L.AuraHeaderBuff, L.AuraHeaderGain, L.AuraHeaderLoss, L.AuraHeaderMusic)
		header:SetFontObject(GameFontNormalSmall)
		box:AddChild(header)

		local s = G:Create('ScrollFrame')
		s:SetLayout('List')
		s:SetRelativeWidth(1)
		s:SetHeight(150)
		box:AddChild(s)

		scroll = s

		PopulateAuraDatabase(self)

		local sort = G:Create('Button')
		sort:SetWidth(140)
		sort:SetText(L.SortListByBuff)
		sort:SetCallback('OnClick', function(w)
			if (sortedBy == 'unit') then
				sortedBy = 'buff'
				w:SetText(L.SortListByUnit)
			else
				sortedBy = 'unit'
				w:SetText(L.SortListByBuff)
			end
			PopulateAuraDatabase(self)
			self:ShowAuraAddEdit(true)
		end)
		f:AddChild(sort)

		local ae = G:Create('InlineGroup')
		ae:SetRelativeWidth(1)
		ae:SetLayout('Flow')
		f:AddChild(ae)
		addEdit = ae

		self:ShowAuraAddEdit(true)
	end

	function ShivvAudio:ShowAuraAddEdit(addMode, unit, buff)
		addEdit:ReleaseChildren()

		self:PopulateMediaListDropdown()

		if (addMode) then
			addEdit:SetTitle(L.AuraAdd)
		else
			addEdit:SetTitle(L.AuraEdit)
		end

		local m = G:Create('Dropdown') -- created here as referenced by following section

		local uBox = G:Create('SimpleGroup')
		uBox:SetLayout('List')
		uBox:SetWidth(140)
		addEdit:AddChild(uBox)

		local uLabel = G:Create('Label')
		uLabel:SetText(L.AuraUnit)
		uBox:AddChild(uLabel)

		local u = G:Create('Dropdown')
		u:SetMultiselect(false)
		u:SetWidth(130)
		u:SetList(unitListDropdown)
		if (addMode) then
			u:SetValue('player')
		else
			u:SetValue(unit)
		end
		u:SetDisabled(not addMode)
		u:SetCallback('OnEnter', function(w)
			ShivvTip(w, L.AuraUnitTip, L.AuraUnitDesc)
		end)
		u:SetCallback('OnLeave', function(w)
			GameTooltip:Hide()
		end)
		u:SetCallback('OnValueChanged', function(w, e, v)
			if (v == 'player') then
				m:SetDisabled(false)
			else
				m:SetDisabled(true)
			end
		end)
		uBox:AddChild(u)

		local onGBox = G:Create('SimpleGroup')
		onGBox:SetLayout('List')
		onGBox:SetWidth(140)
		addEdit:AddChild(onGBox)

		local uLabel = G:Create('Label')
		uLabel:SetText(L.AuraOnGain)
		onGBox:AddChild(uLabel)

		local onG = G:Create('Dropdown')
		onG:SetMultiselect(false)
		onG:SetWidth(130)
		onG:SetList(mediaListDropdown)
		if (addMode) then
			onG:SetValue('none')
		else
			onG:SetValue(ZeroToNone(self.db.profile.auraDB[unit][buff].onGain))
		end
		onG:SetCallback('OnEnter', function(w)
			ShivvTip(w, L.AuraOnGainTip, L.AuraOnGainDesc)
		end)
		onG:SetCallback('OnLeave', function(w)
			GameTooltip:Hide()
		end)
		onGBox:AddChild(onG)

		local onLBox = G:Create('SimpleGroup')
		onLBox:SetLayout('List')
		onLBox:SetWidth(140)
		addEdit:AddChild(onLBox)

		local uLabel = G:Create('Label')
		uLabel:SetText(L.AuraOnLoss)
		onLBox:AddChild(uLabel)

		local onL = G:Create('Dropdown')
		onL:SetMultiselect(false)
		onL:SetWidth(130)
		onL:SetList(mediaListDropdown)
		if (addMode) then
			onL:SetValue('none')
		else
			onL:SetValue(ZeroToNone(self.db.profile.auraDB[unit][buff].onLoss))
		end
		onL:SetCallback('OnEnter', function(w)
			ShivvTip(w, L.AuraOnLossTip, L.AuraOnLossDesc)
		end)
		onL:SetCallback('OnLeave', function(w)
			GameTooltip:Hide()
		end)
		onLBox:AddChild(onL)

		local mBox = G:Create('SimpleGroup')
		mBox:SetLayout('List')
		mBox:SetWidth(140)
		addEdit:AddChild(mBox)

		local uLabel = G:Create('Label')
		uLabel:SetText(L.AuraMusic)
		mBox:AddChild(uLabel)

		-- this item is created above
		m:SetMultiselect(false)
		m:SetWidth(130)
		m:SetList(mediaListDropdown)
		if (addMode) then
			m:SetValue('none')
		else
			m:SetValue(ZeroToNone(self.db.profile.auraDB[unit][buff].music))

			if (unit ~= 'player') then
				m:SetDisabled(true)
			else
				m:SetDisabled(false)
			end
		end
		m:SetCallback('OnEnter', function(w)
			ShivvTip(w, L.AuraMusicTip, L.AuraMusicDesc)
		end)
		m:SetCallback('OnLeave', function(w)
			GameTooltip:Hide()
		end)
		mBox:AddChild(m)

		local b = G:Create('EditBox')
		b:SetLabel(L.AuraBuff)
		b:SetWidth(220)
		if (not addMode) then
			b:SetText(buff)
		end
		b:SetDisabled(not addMode)
		b:SetCallback('OnEnter', function(w)
			ShivvTip(w, L.AuraBuffTip, L.AuraBuffDesc)
		end)
		b:SetCallback('OnLeave', function(w)
			GameTooltip:Hide()
		end)
		addEdit:AddChild(b)

		if (addMode) then
			addEdit:AddChild(Spacer(212, 5))

			local ae = G:Create('Button')
			ae:SetWidth(150)
			ae:SetText(L.AuraAddButton)
			ae:SetCallback('OnClick', function(w)
				if (b:GetText() == '') then
					self:TextOutput(L.AuraEntryMustEnterAura)
				elseif (self.db.profile.auraDB[u:GetValue()] and self.db.profile.auraDB[u:GetValue()][b:GetText()]) then
					self:TextOutput(L.AuraEntryAlreadyExists)
				else
					-- create a new unit entry if needed
					if (not self.db.profile.auraDB[u:GetValue()]) then -- no unit entry yet, make one
						self.db.profile.auraDB[u:GetValue()] = {}
					end

					-- create a new aura entry
					self.db.profile.auraDB[u:GetValue()][b:GetText()] = {
						['onGain'] = NoneToZero(onG:GetValue()),
						['onLoss'] = NoneToZero(onL:GetValue()),
						['music'] = 0,
					}

					-- configure music entry based on whether unit is a player
					if (u:GetValue() == 'player' and m:GetValue() ~= 'none') then
						self.db.profile.auraDB[u:GetValue()][b:GetText()].music = m:GetValue()
					end

					-- got all the values and made an entry... repopulate list and clear fields
					PopulateAuraDatabase(self)

					self:ShowAuraAddEdit(true)

					-- and setup audible auras
					self:AudibleAuras()
				end
			end)
			addEdit:AddChild(ae)
		else
			addEdit:AddChild(Spacer(52, 5))

			local del = G:Create('Button')
			del:SetWidth(150)
			del:SetText(L.AuraDelButton)
			del:SetCallback('OnClick', function(w)
				-- delete the item, clean out the unit entry if its the last buff
				if (self.db.profile.auraDB[unit] and self.db.profile.auraDB[unit][buff]) then -- just check they exist
					self.db.profile.auraDB[unit][buff] = nil

					if (sizeOf(self.db.profile.auraDB[unit]) == 0) then -- this was the last entry for this unit, clear it out
						self.db.profile.auraDB[unit] = nil
					end

					self:TextOutput(L.AuraEntryDeleted)
					PopulateAuraDatabase(self)
					self:ShowAuraAddEdit(true)
				end
			end)
			addEdit:AddChild(del)

			addEdit:AddChild(Spacer(10, 5))

			local ae = G:Create('Button')
			ae:SetWidth(150)
			ae:SetText(L.AuraEditButton)
			ae:SetCallback('OnClick', function(w)
				if (self.db.profile.auraDB[unit] and self.db.profile.auraDB[unit][buff]) then -- just check they exist
					-- can only edit gain/loss/music entries, so only check them

					self.db.profile.auraDB[unit][buff].onGain = NoneToZero(onG:GetValue())
					self.db.profile.auraDB[unit][buff].onLoss = NoneToZero(onL:GetValue())

					if (unit == 'player') then -- is player so set music as well
						self.db.profile.auraDB[unit][buff].music = NoneToZero(m:GetValue())
					end
				end

				self:TextOutput(L.AuraEntryEdited)
				PopulateAuraDatabase(self)
				self:ShowAuraAddEdit(true)
			end)

			addEdit:AddChild(ae)
		end
	end
end

-- ------------ --
-- VITALS PANEL --
-- ------------ --
do
	local vitTabTable = {
		[1] = {text = 'Health I', value = 'h1'},
		[2] = {text = 'Health II', value = 'h2'},
		[3] = {text = 'Health III', value = 'h3'},
		[4] = {text = 'Power I', value = 'p1'},
		[5] = {text = 'Power II', value = 'p2'},
		[6] = {text = 'Power III', value = 'p3'},
	}

	local unitLookup = {
		['h1'] = {'player', 'pet', 'target', 'focus'},
		['h2'] = {'party1', 'party2', 'party3', 'party4'},
		['h3'] = {'arena1', 'arena2', 'arena3', 'arena4', 'arena5'},
		['p1'] = {'player', 'pet', 'target', 'focus'},
		['p2'] = {'party1', 'party2', 'party3', 'party4'},
		['p3'] = {'arena1', 'arena2', 'arena3', 'arena4', 'arena5'},
	}

	local function BuildVitalOption(f, unit, db, self)
		local g = G:Create('SimpleGroup')
		g:SetRelativeWidth(1)
		g:SetLayout('Flow')
		f:AddChild(g)

		local l = G:Create('Label')
		l:SetWidth(80)
		l:SetText(unitListDropdown[unit])
		g:AddChild(l)

		local sBox = G:Create('SimpleGroup')
		sBox:SetLayout('List')
		sBox:SetWidth(150)
		g:AddChild(sBox)

		local sLabel = G:Create('Label')
		sLabel:SetText(L.Sound)
		sBox:AddChild(sLabel)

		local s = G:Create('Dropdown')
		s:SetMultiselect(false)
		s:SetWidth(130)
		s:SetList(mediaListDropdown)
		s:SetValue(ZeroToNone(self.db.profile[db][unit].sound))
		s:SetCallback('OnValueChanged', function(w, e, v)
			if (v == 'none') then
				-- new value is no sound, update module
				self.db.profile[db][unit].sound = 0

				if (db == 'vitalHealthDB') then
					self:MelodicVitals(true, false)
				else
					self:MelodicVitals(false, true)
				end
			elseif (v ~= 'none' and self.db.profile[db][unit].sound == 0) then
				-- new value is a sound but it was no sound, update module
				self.db.profile[db][unit].sound = v

				if (db == 'vitalHealthDB') then
					self:MelodicVitals(true, false)
				else
					self:MelodicVitals(false, true)
				end
			else
				-- any other change just alters played sound, make change
				self.db.profile[db][unit].sound = v
			end
		end)
		sBox:AddChild(s)

		local p = G:Create('Slider')
		p:SetSliderValues(0.05, 0.95, 0.05)
		p:SetIsPercent(true)
		if (db == 'vitalHealthDB') then
			p:SetLabel(L.HealthLevel)
		else
			p:SetLabel(L.PowerLevel)
		end
		p:SetWidth(180)
		p:SetValue(self.db.profile[db][unit].percent)
		p:SetCallback('OnValueChanged', function(w, e, v)
			-- round value out to nearest 0.05 (5%) multiple and save
			self.db.profile[db][unit].percent = (math.floor(math.floor((v * 100) + 0.5) / 5) * 5) / 100
		end)
		g:AddChild(p)

		g:AddChild(Spacer(40, 5))

		local e = G:Create('CheckBox')
		e:SetTriState(false)
		e:SetType('checkbox')
		e:SetLabel(MakeItYellow(L.Enable))
		e:SetDescription('')
		e:SetWidth(80)
		e:SetValue(self.db.profile[db][unit].enable)
		e:SetCallback('OnEnter', function(w)
			ShivvTip(w, L.EnableTip, L.EnableDesc)
		end)
		e:SetCallback('OnLeave', function(w)
			GameTooltip:Hide()
		end)
		e:SetCallback('OnValueChanged', function(w, e, v)
			self.db.profile[db][unit].enable = v

			if (db == 'vitalHealthDB') then
				self:MelodicVitals(true, false)
			else
				self:MelodicVitals(false, true)
			end
		end)
		g:AddChild(e)
	end

	function ShivvAudio:ShowVitals(f)
		local vit = G:Create('TabGroup')
		vit:SetTitle('')
		vit:SetTabs(vitTabTable)
		vit:SetLayout('Flow')
		vit:SetFullHeight(true)
		vit:SetFullWidth(true)
		vit:SetCallback('OnGroupSelected', function(g, e, v)
			g:ReleaseChildren()

			self:ShowVitalsSub(g, v)
		end)
		vit:SelectTab('h1')
		f:AddChild(vit)
	end

	function ShivvAudio:ShowVitalsSub(f, v)
		self:PopulateMediaListDropdown()

		local blurb = G:Create('Label')
		blurb:SetRelativeWidth(1)
		f:AddChild(blurb)

		for _, unit in ipairs(unitLookup[v]) do
			if (strsub(v, 1, 1) == 'h') then
				blurb:SetText(L.HealthBlurb)
				BuildVitalOption(f, unit, 'vitalHealthDB', self)
			else
				blurb:SetText(L.PowerBlurb)
				BuildVitalOption(f, unit, 'vitalPowerDB', self)
			end
		end
	end
end



-- ------------ --
-- EVENTS PANEL --
-- ------------ --
do

	local function BuildEventPanel(f, event, self)
		local box = G:Create('SimpleGroup')
		box:SetLayout('KithTwoPanel')
		box:SetRelativeWidth(0.95)
		box:SetHeight(50)
		f:AddChild(box)

		local label = G:Create('Label')
		label:SetText('\n' .. MakeItYellow(L[event]))
		label:SetFontObject(GameFontNormalSmall)
		box:AddChild(label)

		local sound = G:Create('Dropdown')
		sound:SetMultiselect(false)
		sound:SetWidth(130)
		sound:SetList(mediaListDropdown)
		sound:SetValue(ZeroToNone(self.db.profile.eventDB[event]))
		sound:SetCallback('OnValueChanged', function(w, e, v)
			if ((v == 'none' and self.db.profile.eventDB[event] ~= 0) or (v ~= 'none' and self.db.profile.eventDB[event] == 0)) then
				-- changing from no sound to some sound (or vice versa), update status
				self.db.profile.eventDB[event] = NoneToZero(v)
				self:MusicalEvents()
			else
				self.db.profile.eventDB[event] = NoneToZero(v)
			end
		end)
		box:AddChild(sound)
	end

	function ShivvAudio:ShowEvents(f)
		self:PopulateMediaListDropdown()

		local full = G:Create('SimpleGroup')
		full:SetFullHeight(true)
		full:SetFullWidth(true)
		full:SetLayout('KithTwoPanel')
		f:AddChild(full)

		local sEvent = G:Create('SimpleGroup')
		sEvent:SetRelativeWidth(0.45)
		sEvent:SetFullHeight(true)
		full:AddChild(sEvent)

		local sEventBlurb = G:Create('Label')
		sEventBlurb:SetText(L.SEvent)
		sEventBlurb:SetRelativeWidth(1)
		sEvent:AddChild(sEventBlurb)

		local mEvent = G:Create('SimpleGroup')
		mEvent:SetRelativeWidth(0.45)
		mEvent:SetFullHeight(true)
		full:AddChild(mEvent)

		local mEventBlurb = G:Create('Label')
		mEventBlurb:SetText(L.MEvent)
		mEventBlurb:SetRelativeWidth(1)
		mEvent:AddChild(mEventBlurb)

		BuildEventPanel(sEvent, 'achievement', self)
		BuildEventPanel(sEvent, 'combatstart', self)
		BuildEventPanel(sEvent, 'combatstop', self)
		BuildEventPanel(sEvent, 'death', self)
		BuildEventPanel(sEvent, 'levelup', self)
		BuildEventPanel(mEvent, 'auction', self)
		BuildEventPanel(mEvent, 'bank', self)
		BuildEventPanel(mEvent, 'shop', self)
	end
end

-- ---- --
-- MISC --
-- ---- --
function ShivvAudio:LSMImport()
	local path
	local entriesAdded = 0

	local media = LSM:List('sound')

	for k, name in pairs(media) do
		path = LSM:Fetch('sound', name)

		if (path == '' or path == nil or path == 'None' or path == 'none' or path == 0 or self.db.global.mediaDB[name]) then
			-- do nothing if bad path, or if the entry already exists
		else
			self.db.global.mediaDB[name] = {
				['isMusic'] = false,
				['path'] = path,
			}
			entriesAdded = entriesAdded + 1
		end
	end

	self:TextOutput(entriesAdded .. L.LSMEntriesAdded)
end

function ShivvAudio:TextOutput(text)
	DEFAULT_CHAT_FRAME:AddMessage(L.Title .. ': ' .. text)
end

function ShivvAudio:ChatHandler(input)
	if (input == '') then
		self:BuildGUI() -- open gui if no commands
	elseif (input == 'kill') then
		self:CleanupMusic()
	end
end
