local hpDB, pwrDB = 0, 0
local activeHP, activePWR = {}, {}

function ShivvAudio:MelodicVitals(hp, pwr)
	if (hpDB == 0) then hpDB = self.db.profile.vitalHealthDB end -- setup local references (once) for speed
	if (pwrDB == 0) then pwrDB = self.db.profile.vitalPowerDB end

	self:VitalEventSetup() -- pass repeat timer value to local event section

	-- only do setup for appropriate tags
	if (hp) then
		for unit in pairs(activeHP) do activeHP[unit] = nil end -- clear out tracking data

		if (self.db.profile.vitalEnableHealthTracking) then
			for unit in pairs(hpDB) do
				if (hpDB[unit].enable and hpDB[unit].sound ~= 0) then
					activeHP[unit] = 0 -- setup (and clear) data

					if (UnitExists(unit)) then
						self:UNIT_HEALTH(nil, unit) -- force update if the unit currently exists
					end
				end
			end

			self:RegisterEvent('UNIT_HEALTH') -- register health tracking event
		else
			self:UnregisterEvent('UNIT_HEALTH')
		end
	end

	if (pwr) then
		for unit in pairs(activePWR) do activePWR[unit] = nil end

		if (self.db.profile.vitalEnablePowerTracking) then
			for unit in pairs(pwrDB) do
				if (pwrDB[unit].enable and pwrDB[unit].sound ~= 0) then
					activePWR[unit] = 0 -- setup (and clear) data

					if (UnitExists(unit)) then
						self:UNIT_POWER(nil, unit) -- force update if the unit currently exists
					end
				end
			end

			self:RegisterEvent('UNIT_POWER') -- register power tracking event
		else
			self:UnregisterEvent('UNIT_POWER')
		end
	end
end

do
	local GetTime, UnitHealth, UnitHealthMax = _G.GetTime, _G.UnitHealth, _G.UnitHealthMax -- local speed references
	local hpP, pwrP, hpT, pwrT
	local rTimer

	function ShivvAudio:VitalEventSetup()
		rTimer = self.db.profile.vitalRepeatTimer
	end

	function ShivvAudio:UNIT_HEALTH(event, unit)
		if (not activeHP[unit]) then return end -- abort if not tracking this unit

		hpP = UnitHealth(unit) / UnitHealthMax(unit)

		if (hpP <= hpDB[unit].percent) then	-- health is below threshold, alert!
			hpT = GetTime()

			if ((activeHP[unit] + rTimer) <= hpT and not UnitIsDeadOrGhost(unit)) then -- has it been long enough? and are they alive
				self:SoundHandler(hpDB[unit].sound)
				activeHP[unit] = hpT -- update repeat timer
			end
		end
	end

	function ShivvAudio:UNIT_POWER(event, unit)
		if (not activePWR[unit]) then return end -- abort if not tracking this unit

		pwrP = UnitPower(unit) / UnitPowerMax(unit)

		if (pwrP <= pwrDB[unit].percent) then
			pwrT = GetTime()

			if ((activePWR[unit] + rTimer) <= pwrT and not UnitIsDeadOrGhost(unit)) then
				self:SoundHandler(pwrDB[unit].sound)
				activePWR[unit] = pwrT
			end
		end
	end
end